package XML::Parser;

use strict;
use vars qw($VERSION @ISA);
use Carp;

require DynaLoader;

@ISA = qw(DynaLoader);

$VERSION = "2.06" ;

bootstrap XML::Parser $VERSION;

my %Handler_Setters = (
		       Start => \&SetStartElementHandler,
		       End   => \&SetEndElementHandler,
		       Char  => \&SetCharacterDataHandler,
		       Proc  => \&SetProcessingInstructionHandler,
		       Default => \&SetDefaultHandler,
		       Unparsed => \&SetUnparsedEntityDeclHandler,
		       Notation => \&SetNotationDeclHandler,
		       ExternEnt => \&SetExternalEntityRefHandler,
		       );

sub new {
    my ($class, %args) = @_;
    my $self = bless \%args, $_[0];
    $args{Used} = 0;
    $args{Context} = [];
    $args{Parser} = ParserCreate($args{ProtocolEncoding}, $self);
    $self;
}

sub setHandlers
{
    my ($self, @handler_pairs) = @_;

    croak("Uneven number of arguments to setHandlers method")
	if (int(@handler_pairs) & 1);

    while (@handler_pairs)
    {
	my $type = shift @handler_pairs;
	my $handler = shift @handler_pairs;
	my $hndl = $Handler_Setters{$type};

	unless (defined($hndl))
	{
	    my @types = sort keys %Handler_Setters;

	    croak("Unknown handler type: $type\n Valid types: @types");
	}

	&$hndl($self->{Parser}, $handler);
    }
}

sub default_current
{
    my $self = shift;
    DefaultCurrent($self->{Parser});
}

sub current_line
{
    GetCurrentLineNumber($_[0]->{Parser});
}

sub current_column
{
    GetCurrentColumnNumber($_[0]->{Parser});
}

sub current_byte
{
    GetCurrentByteIndex($_[0]->{Parser});
}

sub base
{
    my ($self, $newbase) = @_;
    my $p = $self->{Parser};
    my $oldbase = GetBase($p);
    SetBase($p, $newbase) if @_ > 1;
    $oldbase;
}

sub context
{
    my $ctx = $_[0]->{Context};
    @$ctx;
}

sub current_element
{
    my ($self) = @_;
    my $ctx = $self->{Context};
    $ {$ctx}[$#{$ctx}];
}

sub in_element
{
    my ($self, $element) = @_;
    my $ctx = $self->{Context};
    $ {$ctx}[$#{$ctx}] eq $element;
}

sub within_element
{
    my ($self, $element) = @_;
    my $cnt = 0;
    foreach (@{$self->{Context}})
    {
	$cnt++ if $_ eq $element;
    }
    return $cnt;
}

sub depth
{
    my ($self) = @_;
    length @{$self->{Context}};
}

sub position_in_context
{
    my ($self, $lines) = @_;
    my $parser = $self->{Parser};
    my ($string, $linepos) = PositionContext($parser, $lines);
    my $col = GetCurrentColumnNumber($parser);
    my $ptr = ('=' x ($col - 1)) . '^' . "\n";
    my $ret;
    my $dosplit = $linepos < length($string);

    $string .= "\n" unless $string =~ /\n$/;

    if ($dosplit)
    {
	$ret = substr($string, 0, $linepos) . $ptr
	    . substr($string, $linepos);
    }
    else
    {
	$ret = $string . $ptr;
    }

    $ret;
}
    
sub DESTROY {
    my $self = shift;
    ParserFree($self->{Parser});
}

sub parse {
    my $self = shift;
    my $arg = shift;
    croak "Parser has already been used" if $self->{Used};
    $self->{Used} = 1;
    my $parser = $self->{Parser};
    my $ioref;
    my $result = 0;
    $self->{ErrorMessage} ||= '';

    if (defined $arg)
    {
	eval {
	    $ioref = *{$arg}{IO};
	};
    }

    if (defined($ioref)) {
      $result = ParseStream($parser, $ioref);
    }
    else {
      $result = ParseString($parser, $arg);
    }

    $result or croak $self->{ErrorMessage};
}

sub parsestring {
    my $self = shift;
    $self->parse(@_);
}

sub parsefile {
    my $self = shift;
    croak "Parser has already been used" if $self->{Used};
    local(*FILE);
    open(FILE, $_[0]) or  croak "Couldn't open $_[0]:\n$!";
    my $ret = $self->parse(*FILE);
    close(FILE);
    $ret;
}

1;

__END__

=head1 NAME

XML::Parser - Access to James Clark's expat XML parser

=head1 SYNOPSIS

 use XML::Parser;

 $parser = new XML::Parser;
 $parser->setHandlers('Start' => \&sh, 'End'   => \&eh,
                      'Default' => \&dh, 'Char' => \&ch);
 open(FOO, 'info.xml') or die "Couldn't open";
 $parser->parse(*FOO);
 close(FOO);
 # $parser->parse('<foo id="me"> here <em>we</em> go </foo>');

 sub sh
 {
     my ($p, $el, %atts) = @_;
     if ($el eq 'special')
     {
	 $p->setHandlers('Default' => \&spec);
	 $p->default_current;
     }
     ...
     if (
 }

 sub spec
 {
     my ($p, $str) = @_;
     ...
     $p->setHandlers('Default' => \&dh);
 }

=head1 DESCRIPTION

This module provides an interface to James Clark's XML parser, expat. As in
expat, a single instance of the parser can only parse one document. Calls
to parsestring after the first for a given instance will die.

Expat (and XML::Parser) are event based. As the parser recognizes parts of
the document (say the start or end of an XML element), then any handlers
registered for that type of an event are called with suitable parameters.

This form of parser can be the basis for building other types, that could,
for instance, create a parse tree or object hierarchy.

=head1 METHODS

=over 4

=item new

This is a class method, the constructor for XML::Parser. Options are
passed as keyword value pairs. The recognized options are:

=over 4

=item * ProtocolEncoding

The protocol encoding name. The default is none. The expat built-in
encodings are: "UTF-8", "ISO-8859-1", "UTF-16", and "US-ASCII". Setting this
overrides any encoding in the XML declaration.

=item * ErrorContext

When this option is defined, errors are reported in context. The value
of ErrorContext should be the number of lines to show on either side of
the line in which the error occurred.

=back

=item setHandlers(TYPE, HANDLER [, TYPE, HANDLER [...]])

This method registers handlers for the various events. If no handlers are
registered, then a call to parsestring or parsefile will only determine if
the corresponding XML document is well formed (by returning without error.)
This may be called from within a handler, after the parse has started.

Setting a handler to something that evaluates to false unsets that
handler.

The recognized events and the parameters passed to the corresponding
handlers are:

=over 4

=item * Start		(Parser, Element [, Attr, Val [,...]])

This event is generated when an XML start tag is recognized. Parser is
an XML::Parser instance. Element is the name of the XML element that is
opened with the start tag. The Attr & Val pairs are generated for each
attribute in the start tag.

=item * End		(Parser, Element)

This event is generated when an XML end tag is recognized. Note that
an XML empty tag (<foo/>) generates both a start and an end event.

There is always a lower level start and end handler installed that wrap
the corresponding callbacks. This is to handle the context mechanism.
A consequence of this is that the default handler (see below) will not
see a start tag or end tag unless the default_current method is called.

=item * Char		(Parser, String)

This event is generated when non-markup is recognized. The non-markup
sequence of characters is in String. A single non-markup sequence of
characters may generate multiple calls to this handler. Whatever the
encoding of the string in the original document, this is given to the
handler in UTF-8.

=item * Proc		(Parser, Target, Data)

This event is generated when a processing instruction is recognized.

=item * Default		(Parser, String)

This is called for any characters that don't have a registered handler.
This includes both characters that are part of markup for which no
events are generated (comments, markup declarations) and characters that
could generate events, but for which no handler has been registered.

Note also that if you install a default handler, then expat doesn't expand
internal entity references.

Whatever the encoding in the original document, the string is returned to
the handler in UTF-8.

=item * Unparsed		(Parser, Entity, Base, Sysid, Pubid, Notation)

This is called for a declaration of an unparsed entity. Entity is the name
of the entity. Base is the base to be used for resolving a relative URI.
Sysid is the system id. Pubid is the public id. Notation is the notation
name. Base and Pubid may be undefined.

=item * Notation		(Parser, Notation, Base, Sysid, Pubid)

This is called for a declaration of notation. Notation is the notation name.
Base is the base to be used for resolving a relative URI. Sysid is the system
id. Pubid is the public id. Base, Sysid, and Pubid may all be undefined.

=item * ExternEnt		(Parser, Base, Sysid, Pubid)

This is called when an external entity is referenced. Base is the base to be
used for resolving a relative URI. Sysid is the system id. Pubid is the public
id. Base, and Pubid may be undefined.

This handler should either return a string, which represents the contents of
the external entity, or return an open filehandle that can be read to obtain
the contents of the external entity, or return undef, which indicates the
external entity couldn't be found and will generate a parse error.

If an open filehandle is returned, it must be returned as either a glob
(*FOO) or as a reference to a glob (e.g. an instance of IO::Handle). The
parser will close the filehandle after using it.

=back

=item default_current

When called from a handler, causes the sequence of characters that generated
the corresponding event to be sent to the default handler (if one is
registered).

=item current_line

Returns the line number of the current position of the parse.

=item current_column

Returns the column number of the current position of the parse.

=item current_byte

Returns the current position of the parse.

=item base([NEWBASE]);

Returns the current value of the base for resolving relative URIs. If
NEWBASE is supplied, changes the base to that value.

=item context

Returns a list of element names that represent open
elements, with the last one being the innermost.

=item current_element

Returns the name of the innermost currently opened element.

=item in_element(NAME)

Returns true if NAME is equal to the name of the innermost currently opened
element.

=item within_element(NAME)

Returns the number of times the given name appears in the context list.

=item depth

Returns the size of the context list.

=item position_in_context(LINES)

Returns a string that shows the current parse position. LINES should be
an integer >= 0 that represents the number of lines on either side of the
current parse line to place into the returned string.

=item parse (SOURCE)

The SOURCE parameter should either be a string containing the whole XML
document, or it should be an open filehandle. Only a single document
may be parsed for a given instance of XML::Parser, so this will croak
if it's been called previously for this instance.

=item parsestring(XML_DOC_STRING)

Parses the given string as an XML document. Only a single document may be
parsed for a given instance of XML::Parser, so this will die if either
parsestring or parsefile has been called for this instance previously.

This method is deprecated in favor of the parse method.

=item parsefile(FILENAME)

Parses the XML document in the given file. Will die if parsestring or
parsefile has been called previously for this instance.

=back

=head1 UNIMPLEMENTED

The following parts of the expat API are not currently accessible via this
extension:

	XML_SetUnknownEncodingHandler

=head1 AUTHORS

Larry Wall <F<larry@wall.org>> wrote version 1.0.
Clark Cooper <F<coopercl@sch.ge.com>> picked up support, changed the API
for this version (2.x), and added some standard package features.

=cut
